@@ -1,24 +1,23 @@
-use crate::storage::{ResourceStorage, Stream};
+use crate::storage::{ResourceStorage, StorageHandle, Stream};
 
 use std::{
-    cell::RefCell,
     collections::BTreeMap,
     fmt,
-    io::{self, Cursor},
+    io::{self, Cursor, Read, Seek, Write},
     path::PathBuf,
-    rc::Rc,
     slice,
+    sync::{Arc, Mutex},
 };
 
-type MemoryStorageStream = Rc<RefCell<Cursor<Vec<u8>>>>;
+type MemoryStorageStream = Arc<Mutex<Cursor<Vec<u8>>>>;
 
 /// Internal storage of data in memory.
 #[derive(Default, Clone)]
 struct MemoryStorage {
     // Streams of resources that were written.
-    streams: Rc<RefCell<BTreeMap<PathBuf, MemoryStorageStream>>>,
+    streams: Arc<Mutex<BTreeMap<PathBuf, MemoryStorageStream>>>,
     // Data of resources that were opened for reading.
-    resources: Rc<RefCell<BTreeMap<PathBuf, Rc<Vec<u8>>>>>,
+    resources: Arc<Mutex<BTreeMap<PathBuf, Arc<Vec<u8>>>>>,
 }
 
 impl fmt::Debug for MemoryStorage {
@@ -27,12 +26,14 @@ impl fmt::Debug for MemoryStorage {
             f,
             "MemoryStorage {{ streams: {:?}, resources: {:?} }}",
             self.streams
-                .borrow()
+                .lock()
+                .unwrap()
                 .iter()
                 .map(|(path, _)| path.display())
                 .collect::<Vec<_>>(),
             self.resources
-                .borrow()
+                .lock()
+                .unwrap()
                 .iter()
                 .map(|(path, _)| path.display())
                 .collect::<Vec<_>>(),
@@ -72,8 +73,8 @@ impl MemoryResourceStorage {
     /// Resources will be placed in ephemeral memory with prefix `path`. A path
     /// has to be provided to unify the interface with `FileResourceStorage`.
     #[allow(clippy::new_ret_no_self)]
-    pub fn new<P: Into<PathBuf>>(path: P) -> Rc<Self> {
-        Rc::new(Self {
+    pub fn new<P: Into<PathBuf>>(path: P) -> Arc<Self> {
+        Arc::new(Self {
             storage: MemoryStorage::default(),
             path: path.into(),
         })
@@ -81,8 +82,8 @@ impl MemoryResourceStorage {
 }
 
 impl ResourceStorage for MemoryResourceStorage {
-    fn subdir(&self, dir: &str) -> Rc<dyn ResourceStorage> {
-        Rc::new(Self {
+    fn subdir(&self, dir: &str) -> StorageHandle {
+        Arc::new(Self {
             storage: self.storage.clone(),
             path: self.path.join(dir),
         })
@@ -90,23 +91,39 @@ impl ResourceStorage for MemoryResourceStorage {
 
     fn exists(&self, resource_name: &str) -> bool {
         let resource_path = self.path.join(resource_name);
-        self.storage.resources.borrow().contains_key(&resource_path)
-            || self.storage.streams.borrow().contains_key(&resource_path)
+        self.storage
+            .resources
+            .lock()
+            .unwrap()
+            .contains_key(&resource_path)
+            || self
+                .storage
+                .streams
+                .lock()
+                .unwrap()
+                .contains_key(&resource_path)
     }
 
     fn read_resource(&self, resource_name: &str) -> Result<&[u8], io::Error> {
         let resource_path = self.path.join(resource_name);
-        if !self.storage.resources.borrow().contains_key(&resource_path) {
-            let streams = self.storage.streams.borrow();
+        if !self
+            .storage
+            .resources
+            .lock()
+            .unwrap()
+            .contains_key(&resource_path)
+        {
+            let streams = self.storage.streams.lock().unwrap();
             let stream = streams.get(&resource_path);
             match stream {
                 Some(stream) => {
                     // Resource is not yet opened, but there is a stream it was written to
                     // => copy the stream as resource data.
-                    let data = Rc::new(stream.borrow().get_ref().clone());
+                    let data = Arc::new(stream.lock().unwrap().get_ref().clone());
                     self.storage
                         .resources
-                        .borrow_mut()
+                        .lock()
+                        .unwrap()
                         .insert(resource_path.clone(), data);
                 }
                 None => {
@@ -117,25 +134,49 @@ impl ResourceStorage for MemoryResourceStorage {
                 }
             }
         }
-        let data = &self.storage.resources.borrow()[&resource_path];
+        let data = &self.storage.resources.lock().unwrap()[&resource_path];
         // We cannot prove to Rust that the buffer will live as long as the storage
         // (we never delete mappings), so we need to manually extend lifetime
         let extended_lifetime_data = unsafe { slice::from_raw_parts(data.as_ptr(), data.len()) };
         Ok(&extended_lifetime_data)
     }
 
-    fn create_output_stream(
-        &self,
-        resource_name: &str,
-    ) -> Result<Rc<RefCell<dyn Stream>>, io::Error> {
+    fn create_output_stream(&self, resource_name: &str) -> Result<Box<dyn Stream>, io::Error> {
         let resource_path = self.path.join(resource_name);
         let stream = self
             .storage
             .streams
-            .borrow_mut()
+            .lock()
+            .unwrap()
             .entry(resource_path)
-            .or_insert_with(|| Rc::new(RefCell::new(Cursor::new(Vec::new()))))
+            .or_insert_with(|| Arc::new(Mutex::new(Cursor::new(Vec::new()))))
             .clone();
-        Ok(stream)
+        Ok(Box::new(StreamWrapper { stream }))
+    }
+}
+
+struct StreamWrapper {
+    stream: Arc<Mutex<Cursor<Vec<u8>>>>,
+}
+
+impl Read for StreamWrapper {
+    fn read(&mut self, result: &mut [u8]) -> std::result::Result<usize, std::io::Error> {
+        self.stream.lock().unwrap().read(result)
+    }
+}
+
+impl Seek for StreamWrapper {
+    fn seek(&mut self, pos: std::io::SeekFrom) -> std::result::Result<u64, std::io::Error> {
+        self.stream.lock().unwrap().seek(pos)
+    }
+}
+
+impl Write for StreamWrapper {
+    fn write(&mut self, data: &[u8]) -> std::result::Result<usize, std::io::Error> {
+        self.stream.lock().unwrap().write(data)
+    }
+
+    fn flush(&mut self) -> std::result::Result<(), std::io::Error> {
+        self.stream.lock().unwrap().flush()
     }
 }
