@@ -1,12 +1,12 @@
 use crate::{
-    arrayview::{debug_format, ArrayView},
     error::ResourceStorageError,
-    structs::{RefFactory, Struct},
+    structs::{NoOverlap, Overlap, Struct},
+    SliceExt,
 };
 
-use crate::{memory, storage::ResourceHandle};
+use crate::storage::ResourceHandle;
 
-use std::{borrow::BorrowMut, fmt, io, marker};
+use std::{borrow::BorrowMut, fmt, io, slice::SliceIndex};
 
 /// A container holding a contiguous sequence of flatdata structs of the same
 /// type `T` in memory, and providing read and write access to it.
@@ -34,23 +34,9 @@ use std::{borrow::BorrowMut, fmt, io, marker};
 /// ```
 ///
 /// ```
-/// # #[macro_use] extern crate flatdata;
-/// # fn main() {
-/// # use flatdata::{ MemoryResourceStorage, Archive, ArchiveBuilder, Vector };
-/// #
-/// # define_struct!(
-/// #     A,
-/// #     RefA,
-/// #     RefMutA,
-/// #     "Schema of A",
-/// #     4,
-/// #     (x, set_x, u32, u32, 0, 16),
-/// #     (y, set_y, u32, u32, 16, 16));
-/// #
-/// # define_archive!(X, XBuilder, "Schema of X";
-/// #     vector(data, false, "Schema of data", set_data, start_data, A),
-/// # );
-/// #
+/// use flatdata::{ MemoryResourceStorage,  Vector };
+/// use flatdata::test::{A, X, XBuilder};
+///
 /// let storage = MemoryResourceStorage::new("/root/extvec");
 /// let builder = XBuilder::new(storage.clone()).expect("failed to create builder");
 /// let mut v: Vector<A> = Vector::new();
@@ -67,66 +53,43 @@ use std::{borrow::BorrowMut, fmt, io, marker};
 ///
 /// let archive = X::open(storage).expect("failed to open");
 /// let view = archive.data();
-/// assert_eq!(view.at(1).x(), 3);
-/// # }
+/// assert_eq!(view[1].x(), 3);
 /// ```
 ///
 /// [`ExternalVector`]: struct.ExternalVector.html
 /// [`as_view`]: #method.as_view
-#[derive(Clone)]
 pub struct Vector<T>
 where
-    T: RefFactory,
+    T: Struct,
 {
-    data: Vec<u8>,
-    _phantom: marker::PhantomData<T>,
+    data: Vec<T>,
 }
 
 impl<T> Vector<T>
 where
-    T: RefFactory,
+    T: Struct,
 {
-    fn padding() -> usize {
-        if <T as Struct>::IS_OVERLAPPING_WITH_NEXT {
-            <T as Struct>::SIZE_IN_BYTES + memory::PADDING_SIZE
-        } else {
-            memory::PADDING_SIZE
-        }
-    }
     /// Creates an empty `Vector<T>`.
     #[inline]
     pub fn new() -> Self {
         Self::with_len(0)
     }
 
+    /// Resets the 'Vector<T>' to its initial empty state.
+    #[inline]
+    pub fn clear(&mut self) {
+        self.data.clear();
+        self.data.push(unsafe { T::create_unchecked() });
+    }
+
     /// Creates a `Vector<T>` with `len` many elements.
     ///
     /// `T`'s fields are all filled with zeroes.
     #[inline]
     pub fn with_len(len: usize) -> Self {
-        let size = Self::calc_size(len);
-        Self {
-            data: vec![0; size],
-            _phantom: marker::PhantomData,
-        }
-    }
-
-    /// Size of the vector in bytes.
-    #[inline]
-    pub fn size_in_bytes(&self) -> usize {
-        self.data.len() - Self::padding()
-    }
-
-    /// Number of elements in the vector.
-    #[inline]
-    pub fn len(&self) -> usize {
-        self.size_in_bytes() / <T as Struct>::SIZE_IN_BYTES
-    }
-
-    /// Returns `true` if the vector has a length 0.
-    #[inline]
-    pub fn is_empty(&self) -> bool {
-        self.len() == 0
+        let mut data = Vec::with_capacity(len + 1);
+        data.resize_with(len + 1, || unsafe { T::create_unchecked() });
+        Self { data }
     }
 
     /// Reserves capacity for at least `additional` more elements to be
@@ -136,57 +99,82 @@ where
     /// if capacity is already sufficient.
     #[inline]
     pub fn reserve(&mut self, additional: usize) {
-        let additional_bytes = Self::calc_size(self.len() + additional) - self.size_in_bytes();
-        self.data.reserve(additional_bytes)
+        self.data.reserve(self.data.len() + additional)
     }
 
-    /// Returns an `ArrayView` to this vector.
+    /// Returns a slice for the vector contents.
+    ///
+    /// Hides sentinels in case T IS_OVERLAPPING_WITH_NEXT,
+    /// i.e. a vector of size 4 get's converted into a slice of size 3
     #[inline]
-    pub fn as_view(&self) -> ArrayView<T> {
-        ArrayView::new(&self.data[..self.size_in_bytes()])
-    }
-
-    /// Returns the contents of this vector as slice of bytes.
-    #[inline]
-    pub fn as_bytes(&self) -> &[u8] {
-        &self.data[..self.size_in_bytes()]
+    pub fn as_view(&self) -> &[T] {
+        if T::IS_OVERLAPPING_WITH_NEXT {
+            &self.data[0..self.data.len().saturating_sub(2)]
+        } else {
+            &self.data[0..self.data.len() - 1]
+        }
     }
 
     /// Appends an element to the end of this vector and returns a mutable
     /// handle to it.
     #[inline]
-    pub fn grow(&mut self) -> <T as Struct>::ItemMut {
-        let old_size = self.data.len();
-        self.data.resize(old_size + <T as Struct>::SIZE_IN_BYTES, 0);
-        let last_index = self.len() - 1;
-        T::create_mut(&mut self.data[last_index * <T as Struct>::SIZE_IN_BYTES..])
+    pub fn grow(&mut self) -> &mut T {
+        let next = self.len();
+        self.data.push(unsafe { T::create_unchecked() });
+        &mut self.data[next]
     }
+}
 
-    /// Return an accessor handle to the element at position `index` in the
-    /// vector.
-    #[inline]
-    pub fn at(&self, index: usize) -> <T as Struct>::Item {
-        T::create(&self.data[index * <T as Struct>::SIZE_IN_BYTES..])
+impl<T> std::ops::Deref for Vector<T>
+where
+    T: Struct,
+{
+    type Target = [T];
+
+    fn deref(&self) -> &[T] {
+        let len = self.data.len() - 1;
+        &self.data[0..len]
     }
+}
 
-    /// Return a mutable handle to the element at position `index` in the
-    /// vector.
-    #[inline]
-    pub fn at_mut(&mut self, index: usize) -> <T as Struct>::ItemMut {
-        T::create_mut(&mut self.data[index * <T as Struct>::SIZE_IN_BYTES..])
+/// Safety: We must not implement DerefMut if T is not NoOverlap,
+/// since it would allow getting a mutable refernce to one element,
+/// and refernece to the next: Both use the same memory address for data of ranges
+impl<T> std::ops::DerefMut for Vector<T>
+where
+    T: Struct + NoOverlap,
+{
+    fn deref_mut(&mut self) -> &mut [T] {
+        let len = self.data.len() - 1;
+        &mut self.data[0..len]
     }
+}
 
-    /// Calculates size in bytes (with padding) needed to store `len` many
-    /// elements.
-    #[inline]
-    fn calc_size(len: usize) -> usize {
-        len * <T as Struct>::SIZE_IN_BYTES + Self::padding()
+impl<T, Idx> std::ops::Index<Idx> for Vector<T>
+where
+    T: Struct + Overlap,
+    Idx: SliceIndex<[T]>,
+{
+    type Output = Idx::Output;
+    fn index(&self, index: Idx) -> &Idx::Output {
+        &self.data[index]
+    }
+}
+
+/// Manually implement IndexMut, since DerefMut is not always implemented
+/// Safety: We must not implement IndexMut for ranges if T is not NoOverlap
+impl<T> std::ops::IndexMut<usize> for Vector<T>
+where
+    T: Struct + Overlap,
+{
+    fn index_mut(&mut self, index: usize) -> &mut T {
+        &mut self.data[index]
     }
 }
 
 impl<T> Default for Vector<T>
 where
-    T: RefFactory,
+    T: Struct,
 {
     /// Creates an empty `Vector<T>`.
     fn default() -> Self {
@@ -194,25 +182,12 @@ where
     }
 }
 
-impl<T> AsRef<[u8]> for Vector<T>
-where
-    T: RefFactory,
-{
-    /// Returns the content of this vector as slice of bytes. Equivalent to
-    /// [`as_bytes`].
-    ///
-    /// [`as_bytes`]: #method.as_bytes
-    fn as_ref(&self) -> &[u8] {
-        self.as_bytes()
-    }
-}
-
 impl<T> fmt::Debug for Vector<T>
 where
-    T: RefFactory,
+    T: Struct + std::fmt::Debug,
 {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        debug_format("Vector", self.as_view().iter(), f)
+        self.as_view().fmt(f)
     }
 }
 
@@ -225,8 +200,7 @@ where
 /// [`grow`] can be accessed and written.
 ///
 /// An external vector *must* be closed, after the last element was written to
-/// it. After closing, it can not be used anymore. Not closing the vector will
-/// result in panic on drop (in debug mode).
+/// it. After closing, it can not be used anymore.
 ///
 /// # Examples
 /// ``` flatdata
@@ -241,23 +215,9 @@ where
 /// ```
 ///
 /// ```
-/// # #[macro_use] extern crate flatdata;
-/// # fn main() {
-/// # use flatdata::{ MemoryResourceStorage, Archive, ArchiveBuilder };
-/// #
-/// # define_struct!(
-/// #     A,
-/// #     RefA,
-/// #     RefMutA,
-/// #     "Schema of A",
-/// #     4,
-/// #     (x, set_x, u32, u32, 0, 16),
-/// #     (y, set_y, u32, u32, 16, 16));
-/// #
-/// # define_archive!(X, XBuilder, "Schema of X";
-/// #     vector(data, false, "Schema of data", set_data, start_data, A),
-/// # );
-/// #
+/// use flatdata::MemoryResourceStorage;
+/// use flatdata::test::{A, X, XBuilder};
+///
 /// let storage = MemoryResourceStorage::new("/root/extvec");
 /// let builder = XBuilder::new(storage.clone()).expect("failed to create builder");
 /// {
@@ -274,40 +234,36 @@ where
 ///
 ///     // data can also be inspected directly after closing
 ///     assert_eq!(view.len(), 2);
-///     assert_eq!(view.at(0).x(), 0);
-///     assert_eq!(view.at(0).y(), 1);
+///     assert_eq!(view[0].x(), 0);
+///     assert_eq!(view[0].y(), 1);
 /// }
 ///
 /// let archive = X::open(storage).expect("failed to open");
 /// let view = archive.data();
-/// assert_eq!(view.at(1).x(), 2);
-/// assert_eq!(view.at(1).y(), 3);
-///
-/// # }
+/// assert_eq!(view[1].x(), 2);
+/// assert_eq!(view[1].y(), 3);
 /// ```
 ///
 /// [`grow`]: #method.grow
 pub struct ExternalVector<'a, T>
 where
-    T: RefFactory,
+    T: Struct,
 {
-    data: Vec<u8>,
+    data: Vector<T>,
     len: usize,
     resource_handle: ResourceHandle<'a>,
-    _phantom: marker::PhantomData<T>,
 }
 
 impl<'a, T> ExternalVector<'a, T>
 where
-    T: RefFactory,
+    T: Struct,
 {
     /// Creates an empty `ExternalVector<T>` in the given resource storage.
     pub fn new(resource_handle: ResourceHandle<'a>) -> Self {
         Self {
-            data: vec![0; memory::PADDING_SIZE],
+            data: Vector::new(),
             len: 0,
             resource_handle,
-            _phantom: marker::PhantomData,
         }
     }
 
@@ -328,46 +284,42 @@ where
     /// may fail due to different IO reasons.
     ///
     /// [`flush`]: #method.flush
-    pub fn grow(&mut self) -> io::Result<<T as Struct>::ItemMut> {
-        if self.data.len() > 1024 * 1024 * 32 {
+    pub fn grow(&mut self) -> io::Result<&mut T> {
+        if self.data.as_view().as_bytes().len() > 1024 * 1024 * 32 {
             self.flush()?;
         }
-        let old_size = self.data.len();
-        self.data.resize(old_size + <T as Struct>::SIZE_IN_BYTES, 0);
         self.len += 1;
-        Ok(T::create_mut(
-            &mut self.data[old_size - memory::PADDING_SIZE..],
-        ))
+        Ok(self.data.grow())
     }
 
     /// Flushes the not yet flushed content in this vector to storage.
     fn flush(&mut self) -> io::Result<()> {
         self.resource_handle
             .borrow_mut()
-            .write(&self.data[..self.data.len() - memory::PADDING_SIZE])?;
-        self.data.resize(0, 0);
-
-        self.data.resize(memory::PADDING_SIZE, 0);
+            .write(&self.data.as_view().as_bytes())?;
+        self.data.clear();
         Ok(())
     }
 
     /// Flushes the remaining not yet flushed elements in this vector and
     /// finalizes the data inside the storage.
     ///
-    /// An external vector *must* be closed, otherwise it will panic on drop
-    pub fn close(mut self) -> Result<ArrayView<'a, T>, ResourceStorageError> {
-        self.flush().map_err(|e| {
-            ResourceStorageError::from_io_error(e, self.resource_handle.name().into())
-        })?;
+    /// An external vector *must* be closed
+    pub fn close(mut self) -> Result<&'a [T], ResourceStorageError> {
+        if self.data.len() > 0 || self.len == 0 {
+            self.flush().map_err(|e| {
+                ResourceStorageError::from_io_error(e, self.resource_handle.name().into())
+            })?;
+        }
         self.resource_handle
             .close()
-            .map(|data| ArrayView::new(data))
+            .map(|data| <&[T]>::from_bytes(data).expect("Corrupted data"))
     }
 }
 
 impl<T> fmt::Debug for ExternalVector<'_, T>
 where
-    T: RefFactory,
+    T: Struct,
 {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         write!(f, "ExternalVector {{ len: {} }}", self.len())
@@ -380,26 +332,7 @@ mod tests {
     // Note: ExternalVector is tested in the corresponding example.
 
     use super::*;
-
-    define_struct!(
-        A,
-        RefA,
-        RefMutA,
-        "no_schema",
-        4,
-        (x, set_x, u32, u32, 0, 16),
-        (y, set_y, u32, u32, 16, 16)
-    );
-
-    define_struct!(
-        R,
-        RefR,
-        RefMutR,
-        "no_schema",
-        4,
-        (first_x, set_first_x, u32, u32, 0, 16),
-        range(x, u32, 0, 16)
-    );
+    use crate::test::{A, R};
 
     #[test]
     fn test_vector_new() {
@@ -410,13 +343,13 @@ mod tests {
     #[test]
     fn test_vector_range() {
         let mut v: Vector<R> = Vector::with_len(3);
-        v.at_mut(0).set_first_x(10);
-        v.at_mut(1).set_first_x(20);
-        v.at_mut(2).set_first_x(30);
+        v[0].set_first_x(10);
+        v[1].set_first_x(20);
+        v[2].set_first_x(30);
 
-        assert_eq!(v.at(0).x(), 10..20);
-        assert_eq!(v.at(1).x(), 20..30);
-        assert_eq!(v.at(2).x(), 30..0);
+        assert_eq!(v[0].x(), 10..20);
+        assert_eq!(v[1].x(), 20..30);
+        assert_eq!(v[2].x(), 30..0);
     }
 
     #[test]
@@ -424,23 +357,23 @@ mod tests {
         let mut v: Vector<A> = Vector::with_len(2);
         assert_eq!(v.len(), 2);
         {
-            let mut a = v.at_mut(0);
+            let a = &mut v[0];
             a.set_x(1);
             a.set_y(2);
             assert_eq!(a.x(), 1);
             assert_eq!(a.y(), 2);
         }
         {
-            let mut b = v.at_mut(1);
+            let b = &mut v[1];
             b.set_x(3);
             b.set_y(4);
             assert_eq!(b.x(), 3);
             assert_eq!(b.y(), 4);
         }
-        let a = v.at(0);
+        let a = &mut v[0];
         assert_eq!(a.x(), 1);
         assert_eq!(a.y(), 2);
-        let b = v.at(1);
+        let b = &mut v[1];
         assert_eq!(b.x(), 3);
         assert_eq!(b.y(), 4);
     }
@@ -450,14 +383,14 @@ mod tests {
         let mut v: Vector<A> = Vector::with_len(1);
         assert_eq!(v.len(), 1);
         {
-            let mut a = v.at_mut(0);
+            let a = &mut v[0];
             a.set_x(1);
             assert_eq!(a.x(), 1);
             a.set_y(2);
             assert_eq!(a.y(), 2);
         }
         let view = v.as_view();
-        let a = view.at(0);
+        let a = &view[0];
         assert_eq!(a.x(), 1);
         assert_eq!(a.y(), 2);
     }
@@ -467,14 +400,14 @@ mod tests {
         let mut v: Vector<A> = Vector::with_len(1);
         assert_eq!(v.len(), 1);
         {
-            let mut a = v.at_mut(0);
+            let a = &mut v[0];
             a.set_x(1);
             a.set_y(2);
             assert_eq!(a.x(), 1);
             assert_eq!(a.y(), 2);
         }
         {
-            let mut b = v.grow();
+            let b = v.grow();
             b.set_x(3);
             b.set_y(4);
             assert_eq!(b.x(), 3);
@@ -482,14 +415,26 @@ mod tests {
         }
         {
             assert_eq!(v.len(), 2);
-            let a = &v.at(0);
+            let a = &v[0];
             assert_eq!(a.x(), 1);
             assert_eq!(a.y(), 2);
-            let b = &v.at(1);
+            let b = &v[1];
             assert_eq!(b.x(), 3);
             assert_eq!(b.y(), 4);
         }
         v.grow();
         assert_eq!(v.len(), 3);
     }
+
+    #[test]
+    fn test_ability_to_get_mut_and_const_for_non_overlap() {
+        let mut v: Vector<A> = Vector::with_len(10);
+        let mut_v_ref = &mut v[..];
+        let (left, right) = mut_v_ref.split_at_mut(5);
+        let x = &mut left[4];
+        let y = &mut right[0];
+        y.set_x(10);
+        assert_eq!(y.x(), 10);
+        assert_eq!(x.x(), 0);
+    }
 }
